import qcore.decorators
from typing import (
    Any,
    Callable,
    Generic,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)
from . import async_task
from . import futures

_T = TypeVar("_T")

def lazy(fn: Callable[..., _T]) -> Callable[..., futures.FutureBase[_T]]: ...
def has_async_fn(fn: object) -> bool: ...
def is_pure_async_fn(fn: object) -> bool: ...
def is_async_fn(fn: object) -> bool: ...
def get_async_fn(
    fn: object, wrap_if_none: bool = ...
) -> Optional[Callable[..., futures.FutureBase[Any]]]: ...
def get_async_or_sync_fn(fn: object) -> Any: ...

class PureAsyncDecoratorBinder(qcore.decorators.DecoratorBinder):
    def is_pure_async_fn(self) -> bool: ...

class PureAsyncDecorator(qcore.decorators.DecoratorBase, Generic[_T]):
    binder_cls = PureAsyncDecoratorBinder
    def __init__(
        self,
        fn: Callable[..., _T],
        task_cls: Optional[Type[futures.FutureBase]],
        kwargs: Mapping[str, Any] = ...,
    ) -> None: ...
    def name(self) -> str: ...
    def is_pure_async_fn(self) -> bool: ...
    def __call__(
        self, *args: Any, **kwargs: Any
    ) -> Union[_T, futures.FutureBase[_T]]: ...
    def __get__(self, owner: Any, cls: Any) -> PureAsyncDecorator[_T]: ...  # type: ignore

class AsyncDecoratorBinder(qcore.decorators.DecoratorBinder, Generic[_T]):
    def asynq(self, *args: Any, **kwargs: Any) -> async_task.AsyncTask[_T]: ...

class AsyncDecorator(PureAsyncDecorator[_T]):
    binder_cls = AsyncDecoratorBinder  # type: ignore
    def is_pure_async_fn(self) -> bool: ...
    def asynq(self, *args: Any, **kwargs: Any) -> async_task.AsyncTask[_T]: ...
    def __call__(self, *args: Any, **kwargs: Any) -> _T: ...
    def __get__(self, owner: Any, cls: Any) -> AsyncDecorator[_T]: ...  # type: ignore

class AsyncAndSyncPairDecoratorBinder(AsyncDecoratorBinder[_T]): ...

class AsyncAndSyncPairDecorator(AsyncDecorator[_T]):
    binder_cls = AsyncAndSyncPairDecoratorBinder  # type: ignore
    def __init__(
        self,
        fn: Callable[..., futures.FutureBase[_T]],
        cls: Optional[Type[futures.FutureBase]],
        sync_fn: Callable[..., _T],
        kwargs: Mapping[str, Any] = ...,
    ) -> None: ...
    def __call__(self, *args: Any, **kwargs: Any) -> _T: ...
    def __get__(self, owner: Any, cls: Any) -> Any: ...

class AsyncProxyDecorator(AsyncDecorator[_T]):
    def __init__(self, fn: Callable[..., futures.FutureBase[_T]]) -> None: ...

class AsyncAndSyncPairProxyDecorator(AsyncProxyDecorator[_T]):
    def __init__(
        self, fn: Callable[..., futures.FutureBase[_T]], sync_fn: Callable[..., _T]
    ) -> None: ...
    def __call__(self, *args: Any, **kwargs: Any) -> _T: ...

class _MkAsyncDecorator:
    def __call__(self, fn: Callable[..., _T]) -> AsyncDecorator[_T]: ...

class _MkPureAsyncDecorator:
    def __call__(self, fn: Callable[..., _T]) -> PureAsyncDecorator[_T]: ...

# In reality these two can return other Decorator subclasses, but that doesn't matter for callers.
@overload
def asynq(  # type: ignore
    *,
    sync_fn: Optional[Callable[..., Any]] = ...,
    cls: Type[futures.FutureBase] = ...,
    **kwargs: Any,
) -> _MkAsyncDecorator: ...
@overload
def asynq(
    pure: bool,
    sync_fn: Optional[Callable[..., Any]] = ...,
    cls: Type[futures.FutureBase] = ...,
    **kwargs: Any,
) -> _MkPureAsyncDecorator: ...  # type: ignore
@overload
def async_proxy(
    *, sync_fn: Optional[Callable[..., Any]] = ...
) -> _MkAsyncDecorator: ...
@overload
def async_proxy(
    pure: bool, sync_fn: Optional[Callable[..., Any]] = ...
) -> _MkPureAsyncDecorator: ...
@asynq()
def async_call(
    fn: Union[Callable[..., _T], Callable[..., futures.FutureBase[_T]]],
    *args: Any,
    **kwargs: Any,
) -> _T: ...
def make_async_decorator(
    fn: Callable[..., Any],
    wrapper_fn: Callable[..., futures.FutureBase[Any]],
    name: str,
) -> Any: ...



import qcore
from typing import Any, Generic, TypeVar

from . import contexts

_T = TypeVar("_T")

class AsyncScopedValue(Generic[_T]):
    def __init__(self, default: _T) -> None: ...
    def get(self) -> _T: ...
    def set(self, value: _T) -> None: ...
    def override(self, value: _T) -> _AsyncScopedValueOverrideContext: ...
    def __call__(self) -> _T: ...

class _AsyncScopedValueOverrideContext(contexts.AsyncContext):
    def __init__(self, target: AsyncScopedValue[_T], value: _T) -> None: ...
    def resume(self) -> None: ...
    def pause(self) -> None: ...

class _AsyncPropertyOverrideContext(contexts.AsyncContext):
    def __init__(self, target: Any, property_name: str, value: Any) -> None: ...
    def resume(self) -> None: ...
    def pause(self) -> None: ...

async_override = _AsyncPropertyOverrideContext



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from asynq import asynq, AsyncContext, scheduler
from asynq.tools import (
    amap,
    afilter,
    afilterfalse,
    amin,
    amax,
    asift,
    asorted,
    acached_per_instance,
    alru_cache,
    alazy_constant,
    aretry,
    call_with_context,
    deduplicate,
    AsyncTimer,
    AsyncEventHook,
)
from qcore.asserts import (
    assert_eq,
    assert_gt,
    assert_is,
    AssertRaises,
    assert_unordered_list_eq,
)
from qcore import get_original_fn
import inspect
import pickle
from unittest import mock


@asynq()
def inner_fn(x):
    pass


@asynq()
def filter_fn(elt):
    yield inner_fn.asynq(elt)
    return elt is not None


@asynq()
def alen(seq):
    return len(seq)


def gen():
    yield 1
    yield None


def test_afilter():
    assert_eq([], afilter.asynq(filter_fn, []).value())
    assert_eq([], afilter.asynq(filter_fn, [None]).value())
    assert_eq([1], afilter(filter_fn, [None, 1, None]))
    assert_eq([1], afilter.asynq(filter_fn, [None, 1, None]).value())
    assert_eq([1], afilter(None, [None, 1, None]))
    assert_eq([1], afilter(filter_fn, gen()))
    assert_eq([1], afilter(None, gen()))


def test_afilterfalse():
    assert_eq([], afilterfalse.asynq(filter_fn, []).value())
    assert_eq([None], afilterfalse.asynq(filter_fn, [None]).value())
    assert_eq([None, None], afilterfalse(filter_fn, [None, 1, None]))
    assert_eq([None, None], afilterfalse.asynq(filter_fn, [None, 1, None]).value())
    assert_eq([None], afilterfalse(filter_fn, gen()))


def test_asift():
    assert_eq(([], []), asift.asynq(filter_fn, []).value())
    assert_eq(([], [None]), asift.asynq(filter_fn, [None]).value())
    assert_eq(([1], [None, None]), asift(filter_fn, [None, 1, None]))
    assert_eq(([1], [None, None]), asift.asynq(filter_fn, [None, 1, None]).value())


def test_amap():
    assert_eq([False], list(amap(filter_fn, [None])))
    assert_eq([True], list(amap(filter_fn, [4])))
    assert_eq([], list(amap(filter_fn, [])))
    assert_eq([False, True, False], list(amap(filter_fn, [None, "", None])))


def test_asorted():
    assert_eq([], asorted([], key=filter_fn))
    assert_eq([None], asorted([None], key=filter_fn))
    assert_eq([None, True], asorted([True, None], key=filter_fn))
    assert_eq([1, 2], asorted([2, 1]))


def test_amax():
    assert_eq(1, amax(1, None, key=filter_fn))
    assert_eq(1, amax([1, None], key=filter_fn))
    assert_eq(1, amax((elt for elt in (1, None)), key=filter_fn))
    assert_eq([1, 2, 3], amax([1], [1, 2, 3], [1, 2], key=alen))
    assert_eq([1, 2, 3], amax([[1], [1, 2, 3], [1, 2]], key=alen))

    assert_eq(4, amax(1, 2, 3, 4))

    with AssertRaises(TypeError):
        amax(key=filter_fn)
    with AssertRaises(ValueError):
        amax([], key=filter_fn)
    with AssertRaises(TypeError):
        amax([], key=filter_fn, random_keyword_argument="raising a TypeError")


def test_amin():
    assert_is(None, amin(1, None, key=filter_fn))
    assert_is(None, amin([1, None], key=filter_fn))
    assert_is(None, amin((elt for elt in (1, None)), key=filter_fn))
    assert_eq([1], amin([1], [1, 2, 3], [1, 2], key=alen))
    assert_eq([1], amin([[1], [1, 2, 3], [1, 2]], key=alen))

    assert_eq(1, amin(1, 2, 3, 4))

    with AssertRaises(TypeError):
        amin(key=filter_fn)
    with AssertRaises(ValueError):
        amin([], key=filter_fn)
    with AssertRaises(TypeError):
        amin([], key=filter_fn, random_keyword_argument="raising a TypeError")


class AsyncObject(object):
    cls_value = 0

    def __init__(self):
        self.value = 0

    @acached_per_instance()
    @asynq()
    def get_value(self, index):
        self.value += 1
        return self.value

    @acached_per_instance()
    @asynq()
    def with_kwargs(self, x=1, y=2, z=3):
        self.value += x + y + z
        return self.value

    @acached_per_instance()
    @asynq()
    def raises_exception(self):
        assert False

    @acached_per_instance()
    @asynq()
    def with_kwonly_arg(self, *, arg=1):
        return arg

    @deduplicate()
    @asynq()
    def increment_value_method(self, val=1):
        self.value += val

    @deduplicate()
    @asynq()
    @staticmethod
    def deduplicated_static_method(val=1):
        AsyncObject.cls_value += val


class UnhashableAcached(AsyncObject):
    __hash__ = None  # type: ignore


def test_acached_per_instance():
    for cls in (AsyncObject, UnhashableAcached):
        obj = cls()
        cache = type(obj).get_value.decorator.__acached_per_instance_cache__
        assert_eq(0, len(cache), extra=repr(cache))

        assert_eq(1, obj.get_value(0))
        assert_eq(1, obj.get_value(0))
        assert_eq(2, obj.get_value(1))
        assert_eq(1, obj.get_value(0))
        assert_eq(1, obj.get_value(index=0))
        assert_eq(1, obj.get_value.asynq(index=0).value())

        assert_eq(8, obj.with_kwargs())
        assert_eq(8, obj.with_kwargs(z=3))
        assert_eq(17, obj.with_kwargs(x=3, y=3))

        assert_eq(1, len(cache), extra=repr(cache))

        assert_eq(1, obj.with_kwonly_arg(arg=1))

        del obj
        assert_eq(0, len(cache), extra=repr(cache))


def test_acached_per_instance_exception_handling():
    obj = AsyncObject()
    try:
        obj.raises_exception()
    except AssertionError:
        # the exception should not affect the internals of the scheduler, and the active task
        # should get cleaned up
        assert_is(None, scheduler.get_active_task())


def test_alru_cache():
    _check_alru_cache()


@asynq()
def _check_alru_cache():
    @alru_cache(maxsize=1, key_fn=lambda args, kwargs: args[0] % 2 == 0)
    @asynq()
    def cube(n):
        return n * n * n

    assert_eq(1, (yield cube.asynq(1)))
    # hit the cache
    assert_eq(1, (yield cube.asynq(3)))
    # cache miss
    assert_eq(8, (yield cube.asynq(2)))
    # now it's a cache miss
    assert_eq(27, (yield cube.asynq(3)))


def test_alazy_constant():
    _check_alazy_constant_no_ttl()
    _check_alazy_constant_ttl()


constant_call_count = 0


@asynq()
def _check_alazy_constant_no_ttl():
    global constant_call_count
    constant_call_count = 0

    @alazy_constant()
    @asynq()
    def constant():
        global constant_call_count
        constant_call_count += 1
        return constant_call_count

    # multiple calls in a short time should only call it once
    assert_eq(1, (yield constant.asynq()))
    assert_eq(1, (yield constant.asynq()))
    assert_eq(1, (yield constant.asynq()))

    # but after a dirty, it should be called again
    constant.dirty()
    assert_eq(2, (yield constant.asynq()))
    assert_eq(2, (yield constant.asynq()))
    assert_eq(2, (yield constant.asynq()))


@asynq()
def _check_alazy_constant_ttl():
    global constant_call_count
    constant_call_count = 0

    @alazy_constant(ttl=100000)  # 100ms
    @asynq()
    def constant():
        global constant_call_count
        constant_call_count += 1
        return constant_call_count

    # multiple calls in a short time should only call it once
    assert_eq(1, (yield constant.asynq()))
    assert_eq(1, (yield constant.asynq()))
    assert_eq(1, (yield constant.asynq()))

    # but after a long enough time, it should be called again
    time.sleep(0.1)
    assert_eq(2, (yield constant.asynq()))
    assert_eq(2, (yield constant.asynq()))
    assert_eq(2, (yield constant.asynq()))

    # or after a dirty
    constant.dirty()
    assert_eq(3, (yield constant.asynq()))
    assert_eq(3, (yield constant.asynq()))
    assert_eq(3, (yield constant.asynq()))


class AnyException(Exception):
    pass


class AnyOtherException(Exception):
    pass


@aretry(Exception)
@asynq()
def retry_it():
    pass


class TestRetry(object):
    def create_function(self, exception_type, max_tries):
        fn_body = mock.Mock()
        fn_body.return_value = []

        @aretry(exception_type, max_tries=max_tries)
        @asynq()
        def function(*args, **kwargs):
            return fn_body(*args, **kwargs)

        return function, fn_body

    def test_pickling(self):
        for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
            pickled = pickle.dumps(retry_it, protocol=protocol)
            assert_is(retry_it, pickle.loads(pickled))

    def test_retry_passes_all_arguments(self):
        function, fn_body = self.create_function(AnyException, max_tries=2)
        function(1, 2, foo=3)
        fn_body.assert_called_once_with(1, 2, foo=3)

    def test_retry_does_not_retry_on_no_exception(self):
        function, fn_body = self.create_function(AnyException, max_tries=3)
        function()
        fn_body.assert_called_once_with()

    def test_retry_does_not_retry_on_unspecified_exception(self):
        function, fn_body = self.create_function(AnyException, max_tries=3)
        fn_body.side_effect = AnyOtherException

        with AssertRaises(AnyOtherException):
            function()

        fn_body.assert_called_once_with()

    def test_retry_retries_on_provided_exception(self):
        max_tries = 4
        function, fn_body = self.create_function(AnyException, max_tries)
        fn_body.side_effect = AnyException

        with AssertRaises(AnyException):
            function()

        assert_eq(max_tries, fn_body.call_count)

    def test_retry_requires_max_try_at_least_one(self):
        with AssertRaises(Exception):
            self.create_function(AnyException, max_tries=0)
        self.create_function(AnyException, max_tries=1)

    def test_retry_can_take_multiple_exceptions(self):
        max_tries = 4

        expected_exceptions = (AnyException, AnyOtherException)

        function, fn_body = self.create_function(expected_exceptions, max_tries)
        fn_body.side_effect = AnyException

        with AssertRaises(AnyException):
            function()

        assert_eq(max_tries, fn_body.call_count)
        fn_body.reset_mock()

        fn_body.side_effect = AnyOtherException

        with AssertRaises(AnyOtherException):
            function()

        assert_eq(max_tries, fn_body.call_count)

    def test_retry_preserves_argspec(self):
        def fn(foo, bar, baz=None, **kwargs):
            pass

        decorated = aretry(Exception)(fn)

        assert_eq(
            inspect.getargspec(fn), inspect.getargspec(get_original_fn(decorated))
        )


class Ctx(AsyncContext):
    is_on = False

    def pause(self):
        Ctx.is_on = False

    def resume(self):
        Ctx.is_on = True


@asynq()
def assert_state(value):
    yield AsyncObject().get_value.asynq(value)
    assert_is(value, Ctx.is_on)


def test_call_with_context():
    assert_state(False)
    call_with_context(Ctx(), assert_state, True)


i = 0


@deduplicate()
@asynq()
def increment_value(val=1):
    global i
    i += val


@deduplicate()
@asynq()
def recursive_incrementer(n):
    if n == 0:
        return (yield increment_value.asynq(n))
    return recursive_incrementer(n - 1)


@deduplicate()
@asynq()
def call_with_dirty():
    call_with_dirty.dirty()


@deduplicate()
@asynq()
def recursive_call_with_dirty():
    global i
    if i > 0:
        return i
    i += 1
    recursive_call_with_dirty.dirty()
    yield recursive_call_with_dirty.asynq()


@asynq()
def dummy():
    pass


@asynq()
def deduplicate_caller():
    yield deduplicated_recusive.asynq()


@deduplicate()
@asynq()
def deduplicated_recusive():
    global i
    existing = i
    i = 1
    yield dummy.asynq()
    if existing == 0:
        deduplicate_caller()


@deduplicate()
@asynq()
def call_with_kwonly_arg(*, arg):
    return arg


def test_deduplicate():
    _check_deduplicate()


@asynq()
def _check_deduplicate():
    global i
    i = 0
    AsyncObject.cls_value = 0

    yield increment_value.asynq()
    assert_eq(1, i)

    yield increment_value.asynq(), increment_value.asynq(1)
    assert_eq(2, i)

    obj = AsyncObject()
    yield obj.increment_value_method.asynq(), obj.increment_value_method.asynq(1)
    assert_eq(1, obj.value)

    yield AsyncObject.deduplicated_static_method.asynq(), AsyncObject.deduplicated_static_method.asynq(
        1
    )
    assert_eq(1, AsyncObject.cls_value)

    i = 0
    yield recursive_call_with_dirty.asynq()

    yield call_with_dirty.asynq()

    with AssertRaises(TypeError):
        yield call_with_kwonly_arg.asynq(1)
    assert_eq(1, (yield call_with_kwonly_arg.asynq(arg=1)))

    i = 0
    deduplicate_caller()


def test_deduplicate_recursion():
    _check_deduplicate_recursion()


@asynq()
def _check_deduplicate_recursion():
    yield recursive_incrementer.asynq(20), increment_value.asynq(0)


def test_async_timer():
    _check_async_timer()


@asynq()
def _slow_task(t):
    yield None
    time.sleep(t)
    return 0


@asynq()
def _timed_slow_task(t):
    with AsyncTimer() as timer:
        yield None
        time.sleep(t)
    return timer.total_time


@asynq()
def _check_async_timer():
    with AsyncTimer() as t:
        results = yield [
            _slow_task.asynq(0.1),
            _timed_slow_task.asynq(0.1),
            _slow_task.asynq(0.1),
            _timed_slow_task.asynq(0.1),
        ]
        assert_eq(0, results[0])
        assert_eq(105000, results[1], tolerance=5000)
        assert_eq(0, results[0])
        assert_eq(105000, results[3], tolerance=5000)

    assert_eq(210000, sum(results), tolerance=10000)
    assert_eq(410000, t.total_time, tolerance=10000)
    assert_gt(t.total_time, sum(results))


def test_async_event_hook():
    calls = []

    @asynq()
    def handler1(*args):
        assert_gt(len(args), 0)
        calls.append("handler1%s" % str(args))

    def handler2(*args):
        calls.append("handler2%s" % str(args))

    hook = AsyncEventHook([handler1])
    hook.subscribe(handler2)

    # trigger
    hook.trigger(1, 2, "a")
    assert_unordered_list_eq(["handler1(1, 2, 'a')", "handler2(1, 2, 'a')"], calls)

    calls = []

    @asynq()
    def async_trigger():
        yield hook.trigger.asynq(2, 3)

    async_trigger()
    assert_unordered_list_eq(["handler1(2, 3)", "handler2(2, 3)"], calls)

    # safe_trigger
    calls = []
    hook2 = AsyncEventHook([handler1, handler2])
    # calling it with no args will raise AssertionError in handler1
    with AssertRaises(AssertionError):
        hook2.safe_trigger()
    assert_eq(["handler2()"], calls)

    # make sure that the order doesn't matter
    calls = []
    hook3 = AsyncEventHook([handler2, handler1])
    # calling it with no args will raise AssertionError in handler1
    with AssertRaises(AssertionError):
        hook3.safe_trigger()
    assert_eq(["handler2()"], calls)


class DeduplicateClassWrapper:
    @deduplicate()
    @asynq()
    def return_three(self):
        return 3

    @deduplicate()
    @asynq()
    def return_five(self):
        return 5

    @asynq()
    def return_three_and_five(self):
        return (yield (self.return_three.asynq(), self.return_five.asynq()))


def test_deduplicate_same_class():
    obj = DeduplicateClassWrapper()

    # make sure the five method has a separate key and therefore there was no cache mixup
    assert_eq((3, 5), obj.return_three_and_five())



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcore.debug import get_bool_by_mask, set_by_mask


class DebugOptions(object):
    """
    All asynq debug options are stored in this structure.

    The class itself is necessary mainly by performance reasons:
    we want to avoid dictionary lookups on checks of debug options.

    """

    def __init__(self):
        self.DUMP_PRE_ERROR_STATE = False
        self.DUMP_EXCEPTIONS = False
        self.DUMP_SCHEDULE_TASK = False
        self.DUMP_CONTINUE_TASK = False
        self.DUMP_SCHEDULE_BATCH = False
        self.DUMP_FLUSH_BATCH = False
        self.DUMP_DEPENDENCIES = False
        self.DUMP_COMPUTED = False
        self.DUMP_NEW_TASKS = False
        self.DUMP_YIELD_RESULTS = False
        self.DUMP_QUEUED_RESULTS = False
        self.DUMP_CONTEXTS = False
        self.DUMP_SYNC = False
        self.DUMP_STACK = False  # When it's meaningful, e.g. on batch flush
        self.DUMP_SCHEDULER_STATE = False
        self.DUMP_SYNC_CALLS = False
        self.COLLECT_PERF_STATS = False

        self.SCHEDULER_STATE_DUMP_INTERVAL = 1  # In seconds
        self.DEBUG_STR_REPR_MAX_LENGTH = 240  # In characters, 0 means infinity
        self.STACK_DUMP_LIMIT = 0  # In frames, None means infinity

        self.ENABLE_COMPLEX_ASSERTIONS = True
        self.KEEP_DEPENDENCIES = False  # don't clear dependencies between yields

    def DUMP_ALL(self, value=None):
        if value is None:
            return get_bool_by_mask(self, "DUMP_")
        set_by_mask(self, "DUMP_", value)


options = DebugOptions()
globals()["options"] = options



from abc import abstractmethod
from typing import ContextManager, Type, Optional, TypeVar
from types import TracebackType
import asynq

from .futures import FutureBase

_T = TypeVar("_T", bound=AsyncContext)

class NonAsyncContext(object):
    def __enter__(self) -> None: ...
    def __exit__(
        self,
        typ: Optional[Type[BaseException]],
        val: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None: ...
    def pause(self) -> None: ...
    def resume(self) -> None: ...

def enter_context(context: ContextManager[object]) -> FutureBase: ...
def leave_context(context: ContextManager[object], active_task: FutureBase) -> None: ...

class AsyncContext(object):
    def __enter__(self: _T) -> _T: ...
    def __exit__(
        self,
        typ: Optional[Type[BaseException]],
        val: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None: ...
    @abstractmethod
    def resume(self) -> None: ...
    @abstractmethod
    def pause(self) -> None: ...



import threading
from typing import Any, List

class LocalProfileState(threading.local): ...

def flush() -> List[Any]: ...
def append(stats: Any) -> None: ...
def reset() -> None: ...
def incr_counter() -> int: ...



from .contexts import AsyncContext
from asynq import asynq, async_proxy

from qcore import Utime
from qcore.events import EventHook
from typing import (
    Any,
    Callable,
    ContextManager,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    overload,
)

_T = TypeVar("_T")
_U = TypeVar("_U")

@asynq()
def amap(function: Callable[[_T], _U], sequence: Iterable[_T]) -> List[_U]: ...
@asynq()
def afilter(
    function: Optional[Callable[[_T], bool]], sequence: Iterable[_T]
) -> List[_T]: ...
@asynq()
def afilterfalse(
    function: Callable[[_T], bool], sequence: Iterable[_T]
) -> List[_T]: ...
@asynq()
def asorted(
    iterable: Iterable[_T],
    key: Optional[Callable[[_T], Any]] = ...,
    reverse: bool = ...,
) -> List[_T]: ...
@overload
@asynq()
def amax(__arg: Iterable[_T], key: Optional[Callable[[_T], Any]] = ...) -> _T: ...
@overload
@asynq()
def amax(*args: _T, key: Optional[Callable[[_T], Any]] = ...) -> _T: ...
@overload
@asynq()
def amin(__arg: Iterable[_T], key: Optional[Callable[[_T], Any]] = ...) -> _T: ...
@overload
@asynq()
def amin(*args: _T, key: Optional[Callable[[_T], Any]] = ...) -> _T: ...
@asynq()
def asift(
    pred: Callable[[_T], bool], items: Iterable[_T]
) -> Tuple[List[_T], List[_T]]: ...

# Used to implement decorators that just wrap a function without changing the signature.
class _AsyncWrapper:
    def __call__(self, fn: _T) -> _T: ...

def acached_per_instance() -> _AsyncWrapper: ...
def alru_cache(
    maxsize: int = ..., key_fn: Optional[Callable[..., Any]] = ...
) -> _AsyncWrapper: ...
def alazy_constant(ttl: int = ...) -> Callable[..., Any]: ...
def aretry(
    exception_cls: type, max_tries: int = ..., sleep: float = ...
) -> Callable[..., Any]: ...
@asynq()
def call_with_context(
    context: ContextManager[Any], fn: Callable[..., _T], *args: Any, **kwargs: Any
) -> _T: ...
def deduplicate(keygetter: Optional[Callable[..., Any]] = ...) -> _AsyncWrapper: ...

class AsyncTimer(AsyncContext):
    total_time: Utime
    def __init__(self) -> None: ...
    def resume(self) -> None: ...
    def pause(self) -> None: ...

class AsyncEventHook(EventHook):
    @asynq()
    def trigger(self, *args: Any) -> None: ...  # type: ignore
    @asynq()
    def safe_trigger(self, *args: Any) -> None: ...  # type: ignore



import qcore
import functools
from typing import Any, Callable, Generic, List, TypeVar

from .decorators import asynq, async_proxy
from .futures import FutureBase
from .utils import result

END_OF_GENERATOR: qcore.MarkerObject

_T = TypeVar("_T")

def async_generator() -> Callable[[Callable[..., Any]], _AsyncGenerator[Any]]: ...

class Value(Generic[_T]):
    def __init__(self, value: _T) -> None: ...

@asynq()
def list_of_generator(generator: _AsyncGenerator[_T]) -> List[_T]: ...
@asynq()
def take_first(generator: _AsyncGenerator[_T], n: int) -> List[_T]: ...

class _AsyncGenerator(Generic[_T]):
    def __init__(self, generator: Any) -> None: ...
    def __iter__(self) -> _AsyncGenerator[_T]: ...
    def next(self) -> FutureBase[_T]: ...
    def __next__(self) -> FutureBase[_T]: ...
    @async_proxy()
    def send(self, value: Any) -> FutureBase[_T]: ...



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import sys
import qcore
import linecache
import traceback
import logging
from sys import stderr, stdout
from pygments import highlight  # pygments loading behavior requires this style import
from pygments import lexers  # see: https://github.com/montylounge/django-sugar/issues/2
from pygments import formatters

from . import _debug

options = _debug.options  # Must be the same object

options.DUMP_PRE_ERROR_STATE = True
options.DUMP_EXCEPTIONS = False
options.DUMP_SCHEDULE_TASK = False
options.DUMP_CONTINUE_TASK = False
options.DUMP_SCHEDULE_BATCH = False
options.DUMP_FLUSH_BATCH = False
options.DUMP_DEPENDENCIES = False
options.DUMP_COMPUTED = False
options.DUMP_NEW_TASKS = False
options.DUMP_YIELD_RESULTS = False
options.DUMP_QUEUED_RESULTS = False
options.DUMP_CONTEXTS = False
options.DUMP_SYNC = False
options.DUMP_STACK = False  # When it's meaningful, e.g. on batch flush
options.DUMP_SCHEDULER_STATE = False
options.DUMP_SYNC_CALLS = False
options.COLLECT_PERF_STATS = False

options.SCHEDULER_STATE_DUMP_INTERVAL = 1  # In seconds
options.DEBUG_STR_REPR_MAX_LENGTH = 240  # In characters, 0 means infinity
options.STACK_DUMP_LIMIT = 10  # In frames, None means infinity
options.MAX_TASK_STACK_SIZE = 1000000  # Max number of concurrent futures + batch items

options.ENABLE_COMPLEX_ASSERTIONS = True
options.KEEP_DEPENDENCIES = False  # don't clear dependencies between yields


def DUMP_ALL(value=None):
    return options.DUMP_ALL(value)


# DUMP_ALL(True)


original_hook = None
is_attached = False
_use_original_exc_handler = False
_should_filter_traceback = True
_use_syntax_highlighting = True
_std_str = str
_std_repr = repr


def enable_original_exc_handler(enable):
    """Enable/disable the original exception handler

    This mainly controls how exceptions are printed when an exception is thrown.
    Asynq overrides the exception handler to better display asynq stacktraces,
    but in some circumstances you may want to show original traces.

    For example, in Jupyter notebooks, the default exception handler displays
    context on exception lines. Enable this function if you'd like that behavior.

    """
    global _use_original_exc_handler
    _use_original_exc_handler = enable


def enable_filter_traceback(enable):
    """Enable/disable replacing asynq boilerplate lines in stacktraces.

    These lines are repeated many times in stacktraces of codebases using asynq.
    By default we replace them so it's easier to read the stacktrace, but you can enable
    if you're debugging an issue where it's useful to know the exact lines.

    """
    global _should_filter_traceback
    _should_filter_traceback = enable


def enable_traceback_syntax_highlight(enable):
    """Enable/disable syntax highlighted stacktraces when using asynq's exception handler."""
    global _use_syntax_highlighting
    _use_syntax_highlighting = enable


def dump_error(error, tb=None):
    """Dumps errors w/async stack traces."""
    try:
        stderr.write("\n" + (format_error(error, tb=tb) or "No error"))
    finally:
        stdout.flush()
        stderr.flush()


def format_error(error, tb=None):
    """Formats errors w/async stack traces."""
    if error is None:
        return None
    result = ""
    if hasattr(error, "_traceback") or tb is not None:
        tb = tb or error._traceback
        tb_list = traceback.format_exception(error.__class__, error, tb)
    elif isinstance(error, BaseException):
        tb_list = traceback.format_exception_only(error.__class__, error)
    else:
        tb_list = []

    tb_text = "".join(tb_list)

    if isinstance(tb_text, bytes):
        tb_text = tb_text.decode("utf-8", "replace")

    if _use_syntax_highlighting:
        tb_text = syntax_highlight_tb(tb_text)

    if _should_filter_traceback:
        # need to do this after syntax highlighting, so we turn it back into a list
        tb_text = "".join(filter_traceback(tb_text.splitlines(True)))

    result += tb_text
    return result


class AsynqStackTracebackFormatter(logging.Formatter):
    """Prints traceback skipping asynq frames during logger.exception/error usages."""

    def formatException(self, exc_info):
        ty, val, tb = exc_info
        return format_error(val, tb=tb)


def _should_skip_frame(frame):
    if frame:
        traceback_hide_directive_name = "__traceback_hide__"
        return (
            frame.f_locals.get(
                traceback_hide_directive_name,
                frame.f_globals.get(traceback_hide_directive_name, False),
            )
            is True
        )
    return False


def extract_tb(tb, limit=None):
    """This implementation is stolen from traceback module but respects __traceback_hide__."""
    if limit is None:
        if hasattr(sys, "tracebacklimit"):
            limit = sys.tracebacklimit
    tb_list = []
    n = 0
    while tb is not None and (limit is None or n < limit):
        f = tb.tb_frame
        if not _should_skip_frame(f):
            lineno = tb.tb_lineno
            co = f.f_code
            filename = co.co_filename
            name = co.co_name
            linecache.checkcache(filename)
            line = linecache.getline(filename, lineno, f.f_globals)
            if line:
                line = line.strip()
            else:
                line = None
            tb_list.append((filename, lineno, name, line))
        tb = tb.tb_next
        n = n + 1
    return tb_list


def format_tb(tb):
    """Formats a traceback into a list of lines."""
    return traceback.format_list(extract_tb(tb))


def dump_stack(skip=0, limit=None):
    """Dumps current stack trace."""
    skip += 2  # To skip dump_stack and traceback.extract_stack
    if limit is None:
        limit = options.STACK_DUMP_LIMIT
    print("--- Stack trace: -----------------------------------------------------")
    try:
        stack = traceback.extract_stack(limit=None if limit is None else limit + skip)
        print("".join(traceback.format_list(stack[:-skip])), end="")
    finally:
        print("----------------------------------------------------------------------")
        stdout.flush()


def dump_asynq_stack():
    """Dumps the current asynq stack to stdout."""
    format_list = format_asynq_stack()
    if format_list is None:
        print("dump_asynq_stack: no asynq task currently active")
    else:
        print("\n".join(format_list))


def format_asynq_stack():
    """Returns a list of strings.

    Each string corresponds to one item in the current asynq stack.

    Returns None if there is no active asynq task.

    """
    # Doing this in the global scope creates a circular dependency
    from .scheduler import get_scheduler

    active_task = get_scheduler().active_task
    if active_task is not None:
        return active_task.traceback()
    else:
        return None


def dump(state):
    if not options.DUMP_PRE_ERROR_STATE:
        return
    stdout.flush()
    stderr.flush()
    stdout.write(
        "\n--- Pre-error state dump: --------------------------------------------\n"
    )
    try:
        state.dump()
    finally:
        stdout.write(
            "----------------------------------------------------------------------\n"
        )
        stderr.write("\n")
        stdout.flush()
        stderr.flush()


def write(text, indent=0):
    if indent > 0:
        indent_str = "  " * indent
        text = text.replace("\n", "\n" + indent_str)
        if not text.startswith("\n"):
            text = indent_str + text
    stdout.write(text + "\n")


def str(source, truncate=True):
    return qcore.safe_str(source, options.DEBUG_STR_REPR_MAX_LENGTH if truncate else 0)


def repr(source, truncate=True):
    return qcore.safe_repr(source, options.DEBUG_STR_REPR_MAX_LENGTH if truncate else 0)


def async_exception_hook(type, error, tb):
    """Exception hook capable of printing async stack traces."""

    stdout.flush()
    stderr.flush()
    if _use_original_exc_handler and original_hook is not None:
        original_hook(type, error, tb)
    dump_error(error, tb=tb)


def ipython_custom_exception_handler(self, etype, value, tb, tb_offset=None):
    """Override ipython's exception handler to print async traceback."""
    async_exception_hook(etype, value, tb)
    # below is the default exception handling behavior of ipython
    if _use_original_exc_handler:
        self.showtraceback()


def attach_exception_hook():
    """Injects async exception hook into the sys.excepthook."""
    try:
        # detect whether we're running in IPython
        __IPYTHON__
    except NameError:
        shell = None
    else:
        # override ipython's exception handler if in a shell.
        # we need to do this because ipython overrides sys.excepthook
        # so just the else block doesn't cover that case.
        from IPython.core.getipython import get_ipython

        # this may be None if __IPYTHON__ is somehow defined, but we are not
        # in fact in a shell
        shell = get_ipython()

    if shell is not None:
        shell.set_custom_exc((BaseException,), ipython_custom_exception_handler)
    else:
        global is_attached, original_hook
        if is_attached:
            sys.stderr.write("Warning: async exception hook was already attached.\n")
            return
        original_hook = sys.excepthook
        sys.excepthook = async_exception_hook
        is_attached = True


def detach_exception_hook():
    """Removes async exception hook into the sys.excepthook."""
    global is_attached, original_hook
    assert is_attached, "Async exception hook wasn't attached."
    sys.excepthook = original_hook
    is_attached = False


@contextlib.contextmanager
def enable_complex_assertions(enable=True):
    old = options.ENABLE_COMPLEX_ASSERTIONS
    try:
        options.ENABLE_COMPLEX_ASSERTIONS = enable
        yield None
    finally:
        options.ENABLE_COMPLEX_ASSERTIONS = old


def disable_complex_assertions():
    return enable_complex_assertions(False)


def sync():
    assert False, (
        "'import asynq' seems broken: this function must be replaced with"
        " asynq.batching.sync."
    )


def get_frame(generator):
    """Given a generator, returns its current frame."""
    if getattr(generator, "gi_frame", None) is not None:
        return generator.gi_frame
    return None


def filter_traceback(tb_list):
    """Given a traceback as a list of strings, looks for common boilerplate and removes it."""

    # List of boiler plate pattern to look for, with example before each
    # The match pattern is just some string that the line needs to contain
    """
        File "asynq/async_task.py", line 169, in asynq.async_task.AsyncTask._continue
        File "asynq/async_task.py", line 237, in asynq.async_task.AsyncTask._continue_on_generator
        File "asynq/async_task.py", line 209, in asynq.async_task.AsyncTask._continue_on_generator
    """
    TASK_CONTINUE = (
        [
            "asynq.async_task.AsyncTask._continue",
            "asynq.async_task.AsyncTask._continue_on_generator",
            "asynq.async_task.AsyncTask._continue_on_generator",
        ],
        "___asynq_continue___",
    )

    """
        File "asynq/decorators.py", line 161, in asynq.decorators.AsyncDecorator.__call__
        File "asynq/futures.py", line 54, in asynq.futures.FutureBase.value
        File "asynq/futures.py", line 63, in asynq.futures.FutureBase.value
        File "asynq/futures.py", line 153, in asynq.futures.FutureBase.raise_if_error
        File "<...>/python3.6/site-packages/qcore/errors.py", line 93, in reraise
           six.reraise(type(error), error, error._traceback)
        File "<...>/python3.6/site-packages/six.py", line 693, in reraise
           raise value
    """
    FUTURE_BASE = (
        [
            "asynq.decorators.AsyncDecorator.__call__",
            "asynq.futures.FutureBase.value",
            "asynq.futures.FutureBase.value",
            "asynq.futures.FutureBase.raise_if_error",
            "reraise",
            "six.reraise",
            "reraise",
            "value",
        ],
        "___asynq_future_raise_if_error___",
    )

    """
        File "asynq/decorators.py", line 153, in asynq.decorators.AsyncDecorator.asynq
        File "asynq/decorators.py", line 203, in asynq.decorators.AsyncProxyDecorator._call_pure
        File "asynq/decorators.py", line 203, in asynq.decorators.AsyncProxyDecorator._call_pure
        File "asynq/decorators.py", line 204, in asynq.decorators.AsyncProxyDecorator._call_pure
        File "asynq/decorators.py", line 275, in asynq.decorators.async_call
    """
    CALL_PURE = (
        [
            "asynq.decorators.AsyncDecorator.asynq",
            "asynq.decorators.AsyncProxyDecorator._call_pure",
            "asynq.decorators.AsyncProxyDecorator._call_pure",
            "asynq.decorators.AsyncProxyDecorator._call_pure",
            "asynq.decorators.async_call",
        ],
        "___asynq_call_pure___",
    )

    REPLACEMENTS = [TASK_CONTINUE, FUTURE_BASE, CALL_PURE]

    # iterate through the lines of the traceback and replace multiline
    # segments that match one of the replacements
    output = []
    i = 0
    while i < len(tb_list):
        did_replacement = False
        # for each replacement, try checking if all lines match
        # if so, replace with the given replacement
        for (text_to_match, replacement) in REPLACEMENTS:
            matches = True
            j = 0
            while j < len(text_to_match) and (i + j) < len(tb_list):
                if text_to_match[j] not in tb_list[i + j]:
                    matches = False
                    break
                j += 1
            if matches and j == len(text_to_match):
                # formatted to match default indentation level.
                output.append("  " + replacement + "\n")
                i = i + j
                did_replacement = True
                break
        if not did_replacement:
            output.append(tb_list[i])
            i += 1
    return output


def syntax_highlight_tb(tb_text):
    """Syntax highlights the traceback so that's a little easier to parse."""
    lexer = lexers.get_lexer_by_name("pytb", stripall=True)
    return highlight(tb_text, lexer, formatters.TerminalFormatter())



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup
from setuptools.extension import Extension

import glob
import os.path


CYTHON_MODULES = [
    "async_task",
    "batching",
    "contexts",
    "_debug",
    "decorators",
    "futures",
    "profiler",
    "scheduler",
    "scoped_value",
    "utils",
]


DATA_FILES = (
    ["py.typed"]
    + ["%s.pxd" % module for module in CYTHON_MODULES]
    + [os.path.relpath(f, "asynq/") for f in glob.glob("asynq/*.pyi")]
)

VERSION = "1.4.0"


EXTENSIONS = [
    Extension("asynq.%s" % module, ["asynq/%s.py" % module])
    for module in CYTHON_MODULES
]


if __name__ == "__main__":
    with open("./README.rst", encoding="utf-8") as f:
        long_description = f.read()

    setup(
        name="asynq",
        version=VERSION,
        author="Quora, Inc.",
        author_email="asynq@quora.com",
        description="Quora's asynq library",
        long_description=long_description,
        long_description_content_type="text/x-rst",
        url="https://github.com/quora/asynq",
        license="Apache Software License",
        classifiers=[
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
        ],
        keywords="quora asynq common utility",
        packages=["asynq", "asynq.tests"],
        package_data={"asynq": DATA_FILES},
        ext_modules=EXTENSIONS,
        setup_requires=["Cython>=0.27.1", "qcore", "setuptools"],
        install_requires=["Cython>=0.27.1", "qcore", "pygments"],
    )



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcore.asserts import assert_eq
from asynq import AsyncScopedValue, asynq, async_override
from asynq.batching import DebugBatchItem

v = AsyncScopedValue("a")


@asynq()
def async_scoped_value_helper(inner_val):
    @asynq()
    def nested():
        assert_eq(v.get(), inner_val)
        yield DebugBatchItem()
        with v.override("c"):
            yield DebugBatchItem()  # just so other function gets scheduled
            assert_eq(v.get(), "c")
            yield DebugBatchItem()

    assert_eq(v.get(), "a")
    yield DebugBatchItem()
    with v.override(inner_val):
        yield DebugBatchItem()
        assert_eq(v.get(), inner_val)
        return (yield nested.asynq())


@asynq()
def async_scoped_value_caller():
    yield async_scoped_value_helper.asynq("e"), async_scoped_value_helper.asynq("f")


def test_async_scoped_value():

    async_scoped_value_caller()

    val = AsyncScopedValue("capybara")
    assert_eq("capybara", val.get())
    val.set("nutria")
    assert_eq("nutria", val.get())

    assert_eq("AsyncScopedValue(nutria)", str(val))
    assert_eq("AsyncScopedValue('nutria')", repr(val))


def test_exception():
    @asynq()
    def test_body():
        assert_eq(v(), "a")
        yield None
        try:
            with v.override("b"):
                yield None
                assert_eq(v(), "b")
                yield None
                raise NotImplementedError()
        except NotImplementedError:
            yield None
            pass
        yield None
        assert_eq(v(), "a")

    test_body()


def test_override():
    class TestObject(object):
        def __init__(self):
            self.v = None

    o = TestObject()
    o.v = "a"

    @asynq()
    def test_body():
        assert_eq(o.v, "a")
        yield None
        with async_override(o, "v", "b"):
            assert_eq(o.v, "b")
            yield None
            try:
                with async_override(o, "v", "c"):
                    assert_eq(o.v, "c")
                    yield None
                    raise NotImplementedError()
            except NotImplementedError:
                pass
            assert_eq(o.v, "b")
        yield None
        assert_eq(o.v, "a")

    test_body()
